package SQL::Interpolate::Macro;
use strict;
use base qw(Exporter);

our @EXPORT;
our %EXPORT_TAGS = (all => [qw(
    relations
    sqland
    sqlif
    sqllink
    sqlor
    sqlparen
    sqlrel
)]);
our @EXPORT_OK = @{$EXPORT_TAGS{'all'}};

#IMPROVE: add--conflict with sql// ?
#sub sql
#{
#    return SQL::Interpolate::SQL->new(@_);
#}

sub sqland
{
    return SQL::Interpolate::And->new(@_);
}

sub sqlor
{
    return SQL::Interpolate::Or->new(@_);
}

sub sqlif
{
    return SQL::Interpolate::If->new(@_);
}

sub sqlrel
{
    return SQL::Interpolate::Rel->new(@_);
}

sub sqllink
{
    return SQL::Interpolate::Link->new(@_);
}

sub sqlparen
{
    return SQL::Interpolate::Paren->new(@_);
}

sub relations
{
    return SQL::Interpolate::RelProcessor->new(@_);
}

# [private]
# Given instances of two relations, generate the SQL to link them.
# For example,
#   ['Sp', ['S', 'p'], $sales_ord_line] and
#   ['p', ['p'], $part]
# gives "Sp.part_nbr = p.part_nbr".
# params:
#   $e1 - entity 1
#   $e2 - entity 2
# where each instance is a arrayref of an entity relation name, a arrayref of
#   names of contained entities, and a relation
#   specification (as passed into C<relations>).
sub _single_link_sql
{
    my($e1, $e2) = @_;
#    die Data::Dumper::Dumper($e1, $e2);
    my($left_idx, $right_idx);
    if   ($e1->[1]->[0] eq  $e2->[1]->[0])
        { $left_idx=0; $right_idx=0; }
    elsif(defined($e2->[1]->[1]) && $e1->[1]->[0] eq  $e2->[1]->[1])
        { $left_idx=0; $right_idx=1; }
    elsif(defined($e1->[1]->[1]) && $e1->[1]->[1] eq  $e2->[1]->[0])
        { $left_idx=1; $right_idx=0; }
    elsif(defined($e1->[1]->[1]) && defined($e2->[1]->[1]) &&
              $e1->[1]->[1] eq  $e2->[1]->[1])
        { $left_idx=1; $right_idx=1; }
    else {
        die "Invalid SQL link [$e1->[0] to $e2->[0]]";
    }

    my $sql = "$e1->[0].$e1->[2]->{key}->[$left_idx ]" . " = " .
              "$e2->[0].$e2->[2]->{key}->[$right_idx]";

    return $sql;
}


1;

package SQL::Interpolate::SQLProcessor;
use strict;
#IMPROVE: package name?

1;

package SQL::Interpolate::RelProcessor;
our @ISA = qw(SQL::Interpolate::SQLProcessor);
use strict;

sub new
{
    my($class, $relations) = @_;
    return bless {
        relations => $relations,
        keys => {}
    }, $class;  #FIX:copy construct?
}

sub init
{
    my $self = shift;
    $self->{keys} = {};
}

sub macro_names
{
    return ('SQL::Interpolate::Rel', 'SQL::Interpolate::Link');
}

sub filter_text
{
    my($self, $sql) = @_;
    while(my($name, $key) = each %{$self->{keys}}) {
        $sql =~ s{ (?<!as\ ) \b \Q$name\E \b (?! \.) }{$key}xge;
    }
    return $sql;
}

sub filter_text_fragment
{
    my($self, $sql, $changing_ref) = @_;
    my @out;
    pos($sql) = 0;
    my $pos0 = pos($sql);
    until($sql =~ /\G$/gc) {
        my $pos1 = pos($sql);
        if($sql =~ m{\G R \( (.*?) \)}xsgc) {
            push @out, substr($sql, $pos0, $pos1 - $pos0) if $pos1 != $pos0;
            $pos0 = pos($sql);
            push @out, SQL::Interpolate::Rel->new($1);
            $$changing_ref = 1;
        }
        elsif($sql =~ m{\G LINK \( (.*?) \)}xsgc) {
            push @out, substr($sql, $pos0, $pos1 - $pos0) if $pos1 != $pos0;
            $pos0 = pos($sql);
            my $params = $1;
            my @params = split /,/, $params;
            s{^\s*|\s*$}{}gs for @params;
            push @out, SQL::Interpolate::Link->new(@params);
            $$changing_ref = 1;
        }
        else {
            $sql =~ m{\G.[^RL]*}xsgc;
        }
    }
    my $pos1 = pos($sql);
    push @out, substr($sql, $pos0, $pos1 - $pos0) if $pos1 != $pos0;
    #die Data::Dumper::Dumper(\@out);
    return @out;
}

1;

package SQL::Interpolate::Rel;
use strict;
our @ISA = qw(SQL::Interpolate::Macro);

sub new
{
    my($class, $name) = @_;

    my $self = bless [
        $name
    ], $class;
    return $self;
}

sub expand
{
    my($self, $dbh, $filter) = @_;  #FIX--generalize interface?

    my $keys = $filter->{keys};
    my $name = $self->[0];
    my $sql;
#    die Data::Dumper::Dumper($filter);
    for my $relation_name (keys %{$filter->{relations}}) {
        my $relation = $filter->{relations}->{$relation_name};
        my $name_re = $relation->{name};

        if($name =~ /^($name_re)$/s) {
            my($name, $name1, $name2) = ($1, $2, $3);

            $keys->{$name1} = "$name.$relation->{key}->[0]";
            $keys->{$name2} = "$name.$relation->{key}->[1]"
                if defined $relation->{key}->[1];
            $sql = "$relation_name as $name";
            last;
        }
    }
    if(! defined $sql) {
        die "Unrecognized relation R($name) in SQL template [$sql].";
    }
    return $sql;
}

1;

package SQL::Interpolate::Link;
use strict;
our @ISA = qw(SQL::Interpolate::Macro);

sub new
{
    my($class, @rels) = @_;

    my $self = bless [
        @rels
    ], $class;
    return $self;
}

sub expand
{
    my($self, $dbh, $filter) = @_;  #FIX--generalize interface?

    #old? my @params = split /,/, $params;  $param =~ s{^\s*|\s*$}{}gs;
    my @params = @$self;

    my $good = 1;
    my $last;
    for my $param (@params) {
        my $match = 0;
        done_param:
        for my $relation (values %{$filter->{relations}}) {
            my $name_re = $relation->{name};
            if($param =~ /^($name_re)/gs) {
                my($name, $name1, $name2) = ($1, $2, $3);
                $param = [$name, [$name1, defined($name2) ? $name2 : ()], $relation];
                $match = 1;
                last done_param;
            }
        }
        if(!$match) {
            die "Invalid param [$param] in LINK macro in SQL template.";
        }
    }

    # relations touching entities.
    my %links;
    my @sql_snips;
    for my $param (@params) {
        for my $entity (@{$param->[1]}) {
            if(defined $links{$entity}) {
                #print Dumper($entity, $links{$entity}, $param), "\n";
                push @sql_snips,
                    SQL::Interpolate::Macro::_single_link_sql($links{$entity}, $param);
            }
        }
        $links{$param->[1]->[0]} = $param;
        $links{$param->[1]->[1]} = $param if defined $param->[1]->[1];
    }

    my $sql = join ' AND ', @sql_snips;

    $sql = "($sql)" if @sql_snips > 1;

    return $sql;
}

1;

package SQL::Interpolate::Paren;
use strict;
our @ISA = qw(SQL::Interpolate::Macro);

sub new
{
    my($class, @elements) = @_;

    my $self = bless [
        @elements
    ], $class;
    return $self;
}

sub expand
{
    my($self, $dbh) = @_;  #FIX--generalize interface?
    return ('(', @$self, ')');
}

1;

package SQL::Interpolate::And;
use strict;
our @ISA = qw(SQL::Interpolate::Macro);

sub new
{
    my($class, @elements) = @_;

    my $self = bless [
        @elements
    ], $class;
    return $self;
}

sub expand
{
    my($self, $dbh) = @_;  #FIX--generalize interface?
    if(@$self == 0) { return '1'; } # trivial case
    else {
        my @out = map {('AND', $_)} @$self;
        shift @out;
        return '(', @out, ')';
    }
}

1;

package SQL::Interpolate::Or;
use strict;
our @ISA = qw(SQL::Interpolate::Macro);

sub new
{
    my($class, @elements) = @_;

    my $self = bless [
        @elements
    ], $class;
    return $self;
}

sub expand
{
    my($self, $dbh) = @_;  #FIX--generalize interface?
    if(@$self == 0) { return '0'; } # trivial case
    else {
        my @out = map {('OR', $_)} @$self;
        shift @out;
        return '(', @out, ')';
    }
}

1;

package SQL::Interpolate::If;
use strict;
our @ISA = qw(SQL::Interpolate::Macro);

sub new
{
    my($class, $condition, $value_if_true) = @_;

    my $self = bless [
        $condition, $value_if_true
    ], $class;
    return $self;
}

sub expand
{
    my($self, $dbh) = @_;  #FIX--generalize interface?
    return $self->[0] ? $self->[1] : ();
}

1;

__END__

=head1 SYNOPSIS

 use SQL::Interpolate qw(:all);
 use SQL::Interpolate::Macro qw(:all);

 # Macros that assist in SQL building
 sql_interp
     q[SELECT * FROM mytable WHERE],
     sqland( sqlif($blue,  q[color = "blue"]),
             sqlif($shape, sqlparen(q[shape =], \$shape)) ),
     q[LIMIT 10]

 # Macros that perform automatic table joining
 # first specify database layout
 my $interp = new SQL::Interpolate(
     relations(
         sales_order      => {name => qr/([S-T])/, key => ['so_nbr']},
         part             => {name => qr/([p-r])/, key => ['part_nbr']},
         sales_order_line => {name => qr/([S-T])([p-r])/,
                              key => ['so_nbr', 'part_nbr']}
     )
 )->make_sql_interp();
 # then do queries
 ($sql, @bind) = $interp->(q[
     SELECT * FROM R(S), R(Sp), R(p)
     WHERE LINK(S,Sp,p) AND S = 123
 ]);
 # Generates SQL
 #   SELECT *
 #   FROM  sales_order as S, sales_order_line as Sp, part as p
 #   WHERE S.so_nbr = Sp.so_nbr AND Sp.part_nbr = p.part_nbr AND
 #         S.so_nbr = 123
  

=head1 DESCRIPTION

This module provides an assortment of macros and SQL filters for
SQL::Interpolate.  These  allow you to write simpler, more robust, and
possibly more portable queries.  You can also write your own macros
and filters.

I<Macros> are objects derived from SQL::Interpolate::Macro.  They can
be inserted into the interpolation list passed to sql_interp().
sql_interp() internally calls sql_flatten(), which expands
any macros to ordinary strings and variable references, which
can then be interpolated directly.  Strings may even contain
macro text (e.g. "WHERE LINK(AB,BC) AND x=y"), which will be
broken down as well (e.g. "WHERE ", link('AB','BC'), " AND x=y")
and then flattened recursively.

An I<SQL filter> is an object derived from
SQL::Interpolate::SQLProcessor.  SQL filters can be used by
sql_interp() and sql_flatten() to assist in macro expansion and/or
filtering of SQL text.  The filtering on SQL text can occur before
and/or after the SQL fragments are interpolated into a single string.

=head2 Motivation

Though a minor point, say you didn't match the parenthesis correctly
in your interpolation list:

  "WHERE thid IN", \@thids,
  "AND ( action IN", \@admin_aids,
         "OR action IN", ['post', 'reply', 'edit']

Typically, this would be caught at runtime rather than (ealier) at
compile time as if you did

"WHERE thid IN", \@thids,
       "AND", paren("action IN", \@admin_aids,
                    "OR action IN", ['post', 'reply', 'edit']
                    # -- syntax error by compiler

Another example may be to handle trivial cases properly, similar to
how SQL::Interpolate handle the special case where the list aftern an
IN is of size zero.  For example,

 my @conditions;  # list of SQL conditional statements.
 ...
 sql_interp "x=2 AND", sql_and(@conditions);

Normally, this would be expanded to

 "x=2 AND", $conditions[0], 'AND', $conditions[1], 'AND', ....

Now, what if @condition == 0?  If we define

 sql_and()  -->  "1"         # analogous to x0 = 1
 sql_or()   -->  "0"         # analogous to x*0 = 0

then proper SQL is generated in the above case:

 "x=2 AND", "1"

which is equivalent to

 "x=2"

=head2 Basic SQL Building Macros

The various macros are described in the Functions sections below.

=head2 Automatic table linking

FIX--Behavior is subject to change after notes from Mark.

FIX--Motivation for this featuer may be recursive data srtuctures.

FIX--change implementation to expand macro with using clause?  Do all
databases support this syntax?

The macro processor can simplify table joins by
aliasing your tables and keys using special naming conventions and
by writing the SQL expressions that link tables together via keys.

This allows one to write

  my $dbx = new DBIx::Interpolate(
      relations(
          sales_order      => {name => qr/([S-T])/, key => ['so_nbr']},
          part             => {name => qr/([p-r])/, key => ['part_nbr']},
          sales_order_line => {name => qr/([S-T])([p-r])/,
                           key => ['so_nbr', 'part_nbr']}
  ));

  $dbx->selectall_arrayref(q[
      SELECT * FROM R(S), R(Sp), R(p)
      WHERE LINK(S,Sp,p) AND S = 123
  ]);

instead of

  $dbh->selectall_arrayref(q[
      SELECT *
      FROM  sales_order        as S,
            sales_order_line   as Sp,
            part               as p
      WHERE S.so_nbr = Sp.so_nbr AND Sp.part_nbr = p.part_nbr AND
            S.so_nbr = 123
  ]);

or

  $dbh->selectall_arrayref(q[
      SELECT *
      FROM  sales_order      as S  JOIN
            sales_order_line as Sp USING (so_nbr) JOIN
            part             as p  USING (part_nbr)
  ]);


The above example prints part information for all the line items on
sales order #123.

The table naming convention is that the names resemble the entities
related by the tables, and this is a natural way to write queries.  In
the above example the entities are sales order (S) and part (p), which
are related by the table represented by the juxtaposition "Sp".
The LINK(S,Sp,p) macro expands by equating the keys in the given relations
that identity the same entities according to the naming convention:

 "S" and "Sp" share S --> "S.so_nbr = Sp.so_nbr"
 "Sp" and "p" share p --> "Sp.part_nbr = p.part_nbr"

Also, any entity written alone in an expression (e.g. "S" in "S = 123" above)
is considered to be shorthand for primary key in a relation representing
that entity:

 "S = 123" --> "S.so_nbr = 123"

The above describes how the naming conventions map to your database
schema.  The meaning of the above is as follows.  sales order
entities can be represented by the names "S" and "T".  Part entities
can be represented by the names "p", "q", and "r".  The juxtaposition
of one of each entity name represents a row in the sales_order_line
relation that relates these two entities.  The entity names in the
matches must be surrounded by capturing parenthesis ().  The names in
the key list correspond respectively to the entities captured in the
match, and these names are the names of the primary or foreign keys
(in the current relation) that identify the entities.  For example,
"Sp" represents a row in the sales_order_line table, with represents a
relationship between a sales order entity (S) and part entity (p).
The key for S is given by Sp.so_nbr, and the key for p is given by
Sp.part_nbr.

It is possible to use two entities in the same class:

 SELECT * FROM R(Sp), R(Tp) WHERE LINK(Sp,Tp)

Also, say the product structure is represented in the database,
where parts can contain (subcomponent) parts, which can in turn contain
other (subcomponent) parts.

 relations(...
     part_part => {
         name => qr/([p-r])([p-r])/, key => ['part_nbr1', 'part_nbr2']}
 )

Now, it is possible to use two entities in the same class in the same
relation:

 SELECT * FROM R(pq), R(qs) WHERE LINK(pq,qs)

Table linking is not limited to SELECT statements either:

 UPDATE R(Sp), R(p)
 SET p.color = 'blue'
 WHERE LINK(Sp,p) AND S = $sonbr

=head2 Writing your own macros

Here is a simple of writing your own macro class.  It is based on the
sqlparen() macro provided by this module.

  sub sqlparan
  {
      return '(', @_, ')';
  }

Example usage:

  my($sql, @bind) = sql_interp "WHERE", sqlparen("x=", \$x);
  # Sets
  #   $sql = "WHERE (x=?)";
  #   @bind = $x;

Here is a more complicated design pattern you may use in case you need to
expand at the time sql_interp is running rather than before.

  # helper function for simpler syntax
  sub sqlparen
  {
      return SQL::Interpolate::Paren->new(@_);
  }
  
  # Macro class
  package SQL::Interpolate::Paren;
  our @ISA = qw(SQL::Interpolate::Macro);
  
  sub new
  {
      my($class, @elements) = @_;

      my $self = bless [
          @elements
      ], $class;
      return $self;
  }
  
  sub expand
  {
      my($self, $dbh, $filter) = @_;  #FIX--generalize interface?
      return ('(', @$self, ')');
  }
  1;

sql_interp will call expand(), which must return the interpolation list
that the macro will expand to.  The interpolation list macro contain
macros itself, an sql_interp will further expand these.

Typically, expansion by sql_interp is useful

(1) if the macro expansion needs data that is available only to an
instance of SQL::Interpolate (e.g. a database handle that is interrogated
to generate database-dependent SQL) or

(2) to support recursive macros properly, e.g.

  sql_interp sqland(sqlor('x=y', 'y=z'), 'z=w')
  # don't expand to
  #   sql_interp sqland('(', 'x=y', 'y=z', ')', 'z=w')
  # which would expand to
  #   sql_interp '(', 'AND', 'x=y', 'AND', 'y=z', 'AND', ')', 'AND', 'z=w'
  # which is not desired.



=head2 Writing your own SQL filters

  # helper function for simpler syntax
  sub sqlparen
  {
      return SQL::Interpolate::Paren->new(@_);
  }
  
  # Macro class
  package SQL::Interpolate::Paren;
  our @ISA = qw(SQL::Interpolate::Macro);
  
  sub new
  {
      my($class, @elements) = @_;

      my $self = bless [
          @elements
      ], $class;
      return $self;
  }
  
  sub expand
  {
      my($self, $dbh, $filter) = @_;  #FIX--generalize interface?
      if(defined($dbh) && $dbh->{Driver}->{Name} eq 'funnydatabase') {
          return 'leftparen', @$self, 'rightparen';
      }
      else {
          return '(', @$self, ')';
      }
  }
  1;

FIX--more documentation.

Your SQL filter may optionally have a few other methods that will be called by
sql_interp if they exist

  # Filter SQL string (after it has been expanded and joined into a single string)
  sub filter_text
  {
      my($self, $sql) = @_;
      ...
      return $sql;
  }

  # Filter a single SQL srting fragment (during expansion)
  sub filter_text_fragment
  {
      my($self, $sql, $changing_ref) = @_;
      if($sql =~ /.../) {
          $$changing_ref = 1;
          ... expand and return list
      }
      else {
          return $sql;
      }
  }

One use of filter_text_fragment is to expand any macros embedded inside strings.

  "SELECT R(AB), R(BC) ..."

expands to

  "SELECT ", r('AB'), ', ', r('BC'), ' ...'

=head2 Functions (that create SQL filter objects)

=over 4

=item C<relations>

Creates an SQL filter for performing automatic table linking.
This creates an SQL::Interpolate::RelProcessor object derived from
SQL::Interpolate::SQLProcessor.

Sets information on relations (tables) for this objects.

 relations({
    messageset => {name => qr/[A-Z]/, key => 'msid'},
    message    => {name => qr/[m-p]/, key => 'mid'},
    messageset_message => {foreign => {
       msid => 'messageset',
       mid  => 'message'
    }},
    messageset_messageset => {foreign => {
       msid_1 => 'messageset',
       msid_2 => 'messageset'
    }},
 });

FIX--more documentation


Expands simplified SQL such as

    SELECT * FROM R(Am), R(m)
    WHERE  A = ? AND LINK(Am, m)

into standard SQL:

    SELECT * FROM messageset_message as Am, message as m
    WHERE  Am.msid = ? AND Am.mid = m.mid

The SQL may contain these macros:

  R(...) - identify relation and entities on it.
     This is converted into an SQL::Interpolate::Rel object.

  LINK(X,Y,Z,...) forms a Boolean expression linking the provided
    relations.  There must be at least one parameter (typically two).
     This is converted into an SQL::Interpolate::Link object.

 $sql_in : string - input string

 $sql_out : string - output string

Possible improvements:

 - support multi-part keys?
 - support optional automatic inclusion of LINK(...).

=back

=head2 Functions (that create macro objects)

=over 4

=item C<sqland>

  $macro = sqland(@predicates);

Creates a macro object (SQL::Interpolate::And) representing
a Boolean AND over a list of predicates.

  ($sql, @bind) = sql_interp sqland('x=y', 'y=z', sqlparen('z=', \3))

Generates

  ('(x=y AND y=z AND 'z=?)', 3)

If the @predicates list is empty, a '1' is returned.

  ($sql, @bind) = sql_interp 'x=y AND', sqland(@predicates);
  # Generates $sql = 'x=y AND 1';

=item C<sqlor>

  $macro = sqlor(@predicates);

Creates a macro object (SQL::Interpolate::Or) representing
a Boolean OR over a list of predicates.

  ($sql, @bind) = sql_interp sqlor('x=y', 'y=z', sqlparen('z=', \3))

Generates

  ('(x=y OR y=z OR 'z=?)', 3)

If the @predicates list is empty, a '0' is returned.

  ($sql, @bind) = sql_interp 'x=y OR', sqlor(@predicates);
  # Generates $sql = 'x=y OR 0';

=item C<sqlparen>

  $macro = sqlparen(@objs)

Creates a macro object (SQL::Interpolate::Paren) representing
a list of interpolation objects to be surrounded by parenthesis.

  ($sql, @bind) = sql_interp sqlparen('x=y AND z=', \3)
  # Generates ($sql, @bind) = ('(x=y and z=?)', 3);

FIX--what if size is zero?

=item C<sqlif>

  $macro = sqlif($condition, $value_if_true)

Creates a macro object (SQL::Interpolate::If) that expands to the given
value if the condition is true, else if expands to the empty list ().

  ($sql, @bind) = sql_interp sqland( sqlif($blue,  q[color = "blue"]),
                                     sqlif($shape, sqlparen(q[shape =], \$shape)) )
  # Generates one of
  #   q[(color = "blue" AND shape = ?)], $shape
  #   q[(color = "blue")]
  #   q[(shape = ?)], $shape
  #   1

=item C<sqlrel>

  $macro = sqlrel($alias)

Creates a macro object (SQL::Interpolate::Rel) that expands to a table and alias
defintion based on the database description given in relations(...).

FIX--more docs

=item C<sqllink>

  $macro = sqllink(@aliases)

Creates a macro object (SQL::Interpolate::Link) that expands to a table join condition
defintion based on the database description given in relations(...).

FIX--more docs

=back

=head1 LIMITATIONS

This module is still under development, so interfaces could
change some, particularly with respect to macros.

=head1 CONTRIBUTORS

David Manura (http://math2.org/david)--author.

Feedback incorporated from Mark Stosberg on table linking, SQL LIMIT, and things.

=head1 LEGAL

Copyright (c) 2004-2005, David Manura.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.
See L<http://www.perl.com/perl/misc/Artistic.html>.

=head1 SEE ALSO

Contained in: L<SQL::Interpolate|SQL::Interpolate>.

Depenencies: L<DBI|DBI>.

Related: L<SQL::Abstract|SQL::Abstract>,
L<DBIx::Abstract|DBIx::Abstract>,
L<Class::DBI|Class::DBI>.

